import { SPoint } from "./SPoint";
import { SSize } from "./SSize";

/**
 * 矩形数据类型定义
 *
 * @author 庞利祥(sybotan@126.com)
 */
export class SRect {
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // 属性定义
    /** 左上角坐标 */
    leftTop: SPoint;
    /** 大小 */
    size: SSize;

    /** 矩形的X轴坐标 */
    get x(): number {
        return this.leftTop.x;
    } // Get x
    set x(value: number) {
        this.leftTop.x = value;
    } // Set x

    /** 矩形的Y轴坐标 */
    get y(): number {
        return this.leftTop.y;
    } // Get x
    set y(value: number) {
        this.leftTop.y = value;
    } // Set y

    /** 矩形的宽 */
    get width(): number {
        return this.size.width;
    } // Get width
    set width(v: number) {
        this.size.width = v;
    } // Set width

    /** 矩形的高 */
    get height(): number {
        return this.size.height;
    } // Get height
    set height(v: number) {
        this.size.height = v;
    } // Set height

    /** 矩形的左 */
    get left(): number {
        return this.leftTop.x;
    } // Get left
    set left(v: number) {
        this.leftTop.x = v;
    } // Set left

    /** 矩形的上 */
    get top(): number {
        return this.y;
    } // Get top
    set top(v: number) {
        this.y = v;
    } // Set top

    /** 矩形的右 */
    get right(): number {
        return this.x + this.width;
    } // get right
    set right(right: number) {
        this.width = right - this.x;
    } // Set right

    /** 矩形的下 */
    get bottom(): number {
        return this.y + this.height;
    } // Get bottom
    set bottom(value: number) {
        this.height = value - this.y;
    } // Set bottom

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // 构造函数
    /**
     * 构造函数
     */
    constructor();

    /**
     *构造函数
     *
     * @param   leftTop         左上角坐标
     * @param   rightBottom     右下角坐标
     */
    constructor(leftTop: SPoint, rightBottom: SPoint);

    /**
     * 构造函数
     *
     * @param   leftTop         左上角坐标
     * @param   size            大小
     */
    constructor(leftTop: SPoint, size: SSize);

    /**
     * 构造函数
     *
     * @param   x           X轴坐标
     * @param   y           Y轴坐标
     * @param   width       宽度
     * @param   height      高度
     */
    constructor(x: number, y: number, width: number, height: number);

    /**
     * 构造函数
     *
     * @param   x           X轴坐标 | 左上角坐标
     * @param   y           Y轴坐标 | 右下角坐标 | 大小
     * @param   width       宽度
     * @param   height      高度
     */
    constructor(
        x?: number | SPoint,
        y?: number | SPoint | SSize,
        width?: number,
        height?: number
    ) {
        if (x == undefined) {
            // constructor()
            this.leftTop = new SPoint(0, 0);
            this.size = new SSize(0, 0);
        } else if (x instanceof SPoint && y instanceof SPoint) {
            // constructor(leftTop: SPoint, rightBottom: SPoint)
            this.leftTop = new SPoint(x.x, x.y);
            this.size = new SSize(y.x - x.x, y.y - x.y);
        } else if (x instanceof SPoint && y instanceof SSize) {
            // constructor(leftTop: SPoint, size: SSize)
            this.leftTop = new SPoint(x.x, x.y);
            this.size = new SSize(y.width, y.height);
        } else {
            // constructor(x: number, y: number, width: number, height: number)
            this.leftTop = new SPoint(x as number, y as number);
            this.size = new SSize(width as number, height as number);
        }
    } // constructor

    /**
     * 是否为空
     *
     * @return  width或height小于等于0，返回true，否则返回false。
     *
     * @see isNull(), isValid()
     */
    isEmpty(): boolean {
        return this.size.isEmpty();
    } // Function isEmpty()

    /**
     * 是否为Null
     *
     * @return  width与height都等于0，返回true，否则返回false。
     *
     * @see isEmpty(), isValid()
     */
    isNull(): boolean {
        return this.size.isNull();
    } // Function isNull()

    /**
     * 是否有效
     *
     * @return  width与height都大于0，返回true，否则返回false。
     * @see isEmpty(), isNull()
     */
    isValid(): boolean {
        return this.size.isValid();
    } // Function isValid()

    /**
     * 是否包含另一个矩阵
     *
     * @return  true    包含
     */
    isIn(rect: SRect): boolean {
        return (
            this.left <= rect.left &&
            this.right >= rect.right &&
            this.bottom >= rect.bottom &&
            this.top <= rect.top
        );
    } // Function isIn()

    /**
     * 判断矩形空间是否包含点x,y
     *
     * @param   x       X坐标
     * @param   y       Y坐标
     * @return  如果包含返回true, 否则返回false
     */
    contains(x: number, y: number): boolean {
        return (
            x >= this.left &&
            x <= this.right &&
            y >= this.top &&
            y <= this.bottom
        );
    } // Function contains()

    /**
     * 计算中心点
     *
     * @return  中心点坐标
     */
    center(): SPoint {
        return new SPoint(
            this.x + this.width / 2.0,
            this.y + this.height / 2.0
        );
    } // Function centerPos()

    /**
     * 平移矩形
     *
     * @param   dx      X轴位移
     * @param   dy      Y轴位移
     */
    translate(dx: number, dy: number): void {
        this.x += dx;
        this.y += dy;
    } // Function translate()

    /**
     * 生成平移矩形
     *
     * @param   dx      X轴位移
     * @param   dy      Y轴位移
     * @return  移动后的矩形
     */
    translated(dx: number, dy: number): SRect {
        return new SRect(this.x + dx, this.y + dy, this.width, this.height);
    } // Function translated()

    /**
     * 调整Rect位置
     *
     * @param   dx          X轴位移
     * @param   dy          Y轴位移
     * @param   dw          宽度调整
     * @param   dh          高度调整
     */
    adjust(dx: number, dy: number, dw: number, dh: number): void {
        this.x += dx;
        this.y += dy;
        this.width += dw;
        this.height += dh;
    } // Function adjusted()

    /**
     * 调整Rect位置
     *
     * @param   dx          X轴位移
     * @param   dy          Y轴位移
     * @param   dw          宽度调整
     * @param   dh          高度调整
     * @return  SRect       调整后的矩形
     */
    adjusted(dx: number, dy: number, dw: number, dh: number): SRect {
        return new SRect(
            this.x + dx,
            this.y + dy,
            this.width + dw,
            this.height + dh
        );
    } // Function adjusted()

    /**
     * 合并矩形
     *
     * @param   rect    合并的矩形
     */
    union(rect: SRect): void {
        let r = this.unioned(rect);
        this.x = r.x;
        this.y = r.y;
        this.width = r.width;
        this.height = r.height;
    } // Function union()

    /**
     * 生成合并矩形
     *
     * @param   rect    合并的矩形
     * @return  返回合并后的矩形
     */
    unioned(rect: SRect): SRect {
        let left = Math.min(this.left, rect.left);
        let top = Math.min(this.top, rect.top);
        let right = Math.max(this.right, rect.right);
        let bottom = Math.max(this.bottom, rect.bottom);
        return new SRect(left, top, right - left, bottom - top);
    } // Function unioned()

    /**
     * 生成相交矩形
     *
     * @param   rect    合并的矩形
     * @return  返回合并后的矩形
     */
    intersected(rect: SRect): SRect {
        let minX = this.left < rect.left ? this.left : rect.left;
        let minY = this.top < rect.top ? this.top : rect.top;
        let maxX = this.right > rect.right ? this.right : rect.right;
        let maxY = this.bottom > rect.bottom ? this.bottom : rect.bottom;
        if (
            this.width + rect.width > maxX - minX &&
            this.height + rect.height > maxY - minY
        ) {
            return new SRect();
        }
        return new SRect();
    } // Function Intersected()

    /**
     * 相交矩形
     *
     * @param   rect    合并的矩形
     */
    intersect(rect: SRect): SRect {
        let left = Math.min(this.left, rect.left);
        let top = Math.min(this.top, rect.top);
        let right = Math.max(this.right, rect.right);
        let bottom = Math.max(this.bottom, rect.bottom);
        return new SRect(left, top, right - left, bottom - top);
    } // Function intersect()
} // Class SRect
